C-----------------------------------------------------------------
C IFFCO - Implicit Filtering For Constrained Optimization.
C
C Code updated 5-18-01 by Alton Patrick.
C MPI version added 12-18-00 by Alton Patrick.
C Revised 11-20-00 for the IBM SP/2 architecture by Alton Patrick.
C Code updated 5-18-99 by Owen Eslinger and Alton Patrick 
C (ojesling@unity.ncsu.edu and hapatric@unity.ncsu.edu).
C Rewritten by Tony Choi (tdchoi@unity.math.ncsu.edu).
C Original version by Paul Gilmore (paul@latour.math.ncsu.edu).
C North Carolina State University
C Dept. of Mathematics
C
C
C IFFCO is a projected quasi-Newton algorithm for constrained
C nonlinear minimization. IFFCO is designed to solve problems of the
C form:
C
C                       min f: Q --> R,
C
C where f is the function to be minimized
C and Q is an n-dimensional hyper-box given by the the following
C equation:
C
C       Q={ x : l(i) <= x(i) <= u(i), i = 1,...,n }.
C
C A brief outline of the algorithm for IFFCO follows.
C
C The main input variables used by IFFCO are:
C       1) An initial point in the hyper-box Q.
C       2) An initial finite difference step (scale) used to calculate
C          the gradient.
C       3) A lower bound for the scales used to calculate the gradient.
C
C IFFCO calculates the function and the gradient of the
C function at the given point. It then tests for convergence
C at the current point with the current scale. If the convergence criteria
C is met, the scale is reduced by a half and the convergence criteria is 
C again checked. This process continues until either the convergence 
C criteria is not satisfied or until the scale is reduced to below the 
C lower bound for scales.
C 
C If the scale has been reduced below the lower bound, the algorithm
C terminates. Otherwise, IFFCO calculates a descent direction 
C and then attempts to find an acceptable new iterate in that direction
C using a cutback line search algorithm.
C 
C If a new point cannot be found, the scale is reduced by half and the
C process continues. If a new point is found, IFFCO continues the process
C of checking for termination and looking for a new acceptable point at
C the current point, scale pair.
C
C IFFCO is designed to solve problems where the function f has the form
C
C                       f(x) = g(x) + o(x).
C
C The function g(x) is a smooth function with a simple form, such as
C a convex quadratic. The function o(x) is a low amplitude high frequency
C function. IFFCO is especially effective on problems where the amplitude
C of o(x) decays near minima of g(x). However, IFFCO has proven to be
C effective on more complex problems.
C
C Input variables:
C       func    -  name of subroutine that evaluates f(x).
C                  Declare it external within the routine that calls
C                  IFFCO.  e.g. external myfunction.
C       x       -  (dble/size n) initial vector.
C       u       -  (dble/size n) vector of upper constraints.
C       l       -  (dble/size n) vector of lower constraints.
C       fscale  -  (dble) approximate maximum value of the function
C                  within the constraints.
C       minh    -  (dble) lower bound for the scale (h).
C       maxh    -  (dble) upper bound for the scale (h).
C       n       -  (int) the dimension of x.
C       maxit(1)-  (int) maximum iterations on a given scale.
C       maxit(2)-  (int) maximum func. evals for entire run.
C       restart -  (int) # of restarts to be done.
C       writ    -  (int) type of output (see statsIF routine).
C                    writ = 0 - suppress all output
C                    writ = 1 - standard info
C                    writ = 2 - standard info + x + #f evals
C                    writ = 3 - standard info, unscaled
C                    writ = 4 - standard info + x + #f evals, unscaled
C                  Add 10 to writ to print output to the screen.
C       termtol -  (dble) tolerence at which to terminate the iteration
C                  for a given scale.
C       maxcuts -  (int) maximum # of cuts allowed in line search.
C       option  -  (int/size 7) vector 
C                    option(1) = Quasi-Newton update.
C                              = 0 (Gradient projection)
C                              = 1 (SR1)
C                              = 2 (BFGS)
C                    option(2) = minimum strategy.
C                              = 0 (Do nothing)
C                              = 1 (Take min as current point at restart)
C                              = 2 (Take min as current point at new scale)
C                              = 3 (Take min as current point at new step)
C                    option(3) = initial point.
C                              = 0 (Use given initial point)
C                              = 1 (Use center of box as initial point)
C                    option(4) = Quasi-Newton variation.
C                              = 0 (Re-initialize B at each new scale)
C                              = 1 (Re-initialize B if active set changes)
C                              = 2 (Update Hessian even if active set changes)
C                    option(5) = Output file
C                              = 0 (Create and use iffco.out (default))
C                              = 1 (option(6) contains the unit number of the
C                                  file to use instead of iffco.out)
C                    option(6) = Output file unit number.
C                    option(7) = Master processor use.
C                              = 0 (Use master processor for function
C                                  evaluations when computing the gradient
C                                  (default))
C                              = 1 (Do not use master processor for function 
C                                  evaluations when computing the gradient)
C       idata   -  (int/size ilen) vector of int's that is passed, unaltered,
C                    to the user's objective function code.
C       ilen    -  (int) length of vector idata.
C       ddata   -  (dble/size dlen) vector of dble's that is passed, unaltered,
C                    to the user's objective function code.
C       dlen    -  (int) length of vector ddata.
C       cdata   -  (char/size clen) vector of char's that is passed, unaltered,
C                    to the user's objective function code.
C
C Output variables:
C       x       - IFFCO's candidate for the solution.
C       f       - (dble) f(x).
C
C Global variables:
C   Global variables are only used to keep track of the current min and 
C   max of all sampled points, total number of function evaluations, and 
C   machine precision.  Suffixed w/ 'IF' for less chance of duplication with
C   other globals.  
C   Found in: iffco, minIF, funcIF, initIF, statsIF, par_gradIF, 
C   ser_linesearchIF, par_linesearchIF, maxIF, quasiIF, restartIF, takeminIF.
C
C Since FORTRAN does not allow local variable arrays to be dynamic you must 
C change the parameter mx in every routine if n > mx.  Another parameter
C you may have to change in many routines is maxprocs.  maxprocs should
C be >= nprocs.  It is currently set to mx*2+1, which should be large enough
C unless you allow very many cutbacks in the linesearch.
C
C fhist (dble/size mx2) is a local array that stores f(x) for every iteration.  
C We store the unscaled value since fscale can change after restarts.  
C Parameter mx2 determines the size of fhist.  If you expect more then you 
C must change mx2.
C
C Subroutines:
C Note all subroutines are suffixed w/ 'IF' (for IFFCO).
C       1)  inputsIF
C       2)  initIF
C       3)  statsIF
C       4)  scaleIF
C       5)  unscaleIF
C       6)  funcIF
C       7a) ser_gradIF
C       7b) par_gradIF
C       8)  minIF
C       9)  maxIF
C       10) updateIF
C       11a)ser_linesearchIF
C       11b)par_linesearchIF
C       12) eyeIF
C       13) evaltolIF
C       14) quasiIF
C       15) stepIF
C       16) restartIF
C       17) takeminIF
C       18) pointsIF
C Parallel communication subroutines (only used by the MPI and PVM versions):
C       c1) mastersendIF
C       c2) slaverecvIF
C       c3) slavesendIF
C       c4) masterrecvIF
C       c5) getmytidIF
C       c6) getnprocsIF
C       c7) gettidIF
C       c8) comminitIF
C       c9) commexitIF
C-----------------------------------------------------------------
      subroutine iffco(func,x,u,l,fscale,minh,maxh,n,maxit,
     *     restart,writ,termtol,f,maxcuts,option,
     *     idata, ilen, ddata, dlen, cdata, clen)

        implicit none

C Arguments
        integer n, maxit(2), restart, writ, maxcuts, option(7)
        double precision x(n), u(n), l(n)
        double precision fscale, minh, maxh, termtol, f 
        external func
        integer ilen, dlen, clen
        integer idata(ilen) 
        double precision ddata(dlen)
        character cdata(clen)

C Local variables
        integer mx, mx2
        parameter(mx = 24)
        parameter(mx2 = 1000)
        integer i, k1, its, ncuts, quasi, flag, fflag, fctr
        integer nprocs, mainout_unum, ptsout_unum
        double precision tol, x_orig(mx), xp(mx), fp
        double precision B(mx,mx), g(mx), gp(mx), p(mx)
        double precision h, fhist(mx2)
        integer stencil

C Global variables
        integer nevalsIF
        double precision fminIF, xminIF(mx), mepsIF, fmaxIF
        common /globalIF/nevalsIF
        common /globalIF2/fminIF,xminIF,fmaxIF
        common /globalIF3/mepsIF



C INITIALIZE IFFCO AND CHECK FOR ERRORS. TERMINATE IF ANY ARE FOUND.
C
C

C       Set unit number of files that might be written to (if writ > 10).
C       The main output file
        mainout_unum = 31
C       The record of points evaluated
        ptsout_unum = 32
        

        nprocs = 1

        
C       Report the input parameters.
        call inputsIF(n,x,u,l,fscale,minh,maxh,maxit,
     *     restart,writ,termtol,f,maxcuts,option,nprocs,
     *     mainout_unum,ptsout_unum)

C       Initialize, error check parameters, and copy x to x_orig.
        call initIF(n,mx,x,u,l,func,fscale,minh,maxh,maxit,restart,
     *       writ,termtol,maxcuts,option,x_orig,h,quasi,flag,
     *       mainout_unum)

C       Any errors?
        if(flag.gt.0) then

          return
        end if

C-----------------------------------------

C BEGIN IFFCO BY SETTING X_ORIG AND FINDING INITIAL F.
C
C
C       Scale x_orig onto the unit box.  Routine returns x.
        call scaleIF(n,x_orig,u,l,x)

C       Evaluate function at initial point, x.
	    call funcIF(n,x,u,l,func,fscale,f,fflag,writ,
     *    mainout_unum,ptsout_unum,
     *    idata, ilen, ddata, dlen, cdata, clen)

C       Initialize fhist(unscaled), fminIF(unscaled),
C       xminIF(unscaled), fmaxIF (scaled) to initial point.
        fhist(1) = f*fscale
        fctr = 2

        fminIF = f*fscale
        fmaxIF = f
        do 30 i=1,n
          xminIF(i) = x_orig(i)
 30     continue



C Loop over restarts.
        do 40 k1=1,(restart + 1)

C         Initialize quasi-Newton matrix.
          call eyeIF(n,B)


C Loop over scales.
          do while (h.ge.minh .and. nevalsIF.lt.maxit(2))

C           Re-initialize quasi-Newton matrix if appropriate.
            if (option(4).eq.0) then
              call eyeIF(n,B)
            end if

            its = 0
            ncuts = 0

C           Evaluate gradient.

            call ser_gradIF(n, x, u, l, func, fscale, f, h,
     *        stencil, g, writ, mainout_unum, ptsout_unum,
     *        idata, ilen, ddata, dlen, cdata, clen)


C           Evaluate tolerance.
            call evaltolIF(n,x,u,l,fscale,g,tol)


C Iterate until: 1) maxit exceeded.
C             or 2) iteration has converged for this scale.
C             or 3) line search failure.
C             or 4) Stencil failure

            do while (tol.gt.termtol*h .and. stencil.eq.0 .and.
     *                ncuts.ge.0 .and. its.lt.maxit(1)
     *                .and. nevalsIF.lt.maxit(2))

C             Form step: p = -H*g.
              call stepIF(n,g,h,B,quasi,writ,mainout_unum,p)

C             Line Search.

              call ser_linesearchIF(n,x,u,l,func,fscale,g,p,f,minh,
     *          maxcuts,ncuts,fp,xp,writ,mainout_unum,ptsout_unum,
     *          idata, ilen, ddata, dlen, cdata, clen)


C             Take minimum if appropriate.
              if (option(2).ge.3) then
                  call takeminIF(n,u,l,mx2,fscale,writ,mainout_unum,xp,
     *              fp,fhist,fctr,flag)

C               Set ncuts = 0 if we take the min.
                if (flag.eq.1.and.ncuts.lt.0) then
                  ncuts = 0
                end if
              end if

C             Accept or reject new point (accept if ncuts >= 0).
              if(ncuts.ge.0) then
C               Get new gradient.

                call ser_gradIF(n,xp,u,l,func,fscale,fp,h,stencil,gp,
     *            writ, mainout_unum, ptsout_unum,
     *            idata, ilen, ddata, dlen, cdata, clen)



C               Quasi-Newton Update.
                call quasiIF(n,x,xp,g,gp,quasi,option(4),B,
     *            writ, mainout_unum)

C               Update x,g,f and increment iteration counter.
                call updateIF(n,xp,gp,fp,x,g,f)
                its = its + 1

C               Report Stats (new x,f,g).
                call statsIF(n,x,u,l,fscale,h,f,writ,ncuts,g,its,
     *            mainout_unum)

C               Save f history (unscaled).
                fhist(fctr) = f*fscale
                fctr = fctr + 1

C               Evaluate tolerance.
                call evaltolIF(n,x,u,l,fscale,g,tol)
              end if

            end do

C           Check if exited loop because of convergence or
C           exceeding maxit.
            if(tol.le.termtol*h) then
              ncuts = -2
            else if(its.ge.maxit(1)) then
              ncuts = -3
            else if(stencil.eq.1) then
              ncuts = -4
            else if(nevalsIF.ge.maxit(2)) then
              ncuts = -5
            end if

C           Report Stats (new x,f,g).
            call statsIF(n,x,u,l,fscale,h,f,writ,ncuts,g,its,
     *        mainout_unum)

C           Save f history (unscaled).
            fhist(fctr) = f*fscale
            fctr = fctr + 1

C           Take minimum if appropriate.
            if (option(2).ge.2) then
              call takeminIF(n,u,l,mx2,fscale,writ,mainout_unum,x,
     *          f,fhist,fctr,flag)
            end if

C           Set h-scale value.
            h = h / 2D0

          end do

C         Perform restart if there is a restart to be done.
          if(k1.lt.restart+1) then
            call restartIF(f,maxh,writ,mainout_unum,fscale,h)

C           Take minimum if appropriate.
            if (option(2).ge.1) then
              call takeminIF(n,u,l,mx2,fscale,writ,mainout_unum,x,
     *                       f,fhist,fctr,flag)
            end if
          end if

 40     continue

C       Perform concluding rites.

C       Take minimum if appropriate.
        if (option(2).ge.1) then
          call takeminIF(n,u,l,mx2,fscale,writ,mainout_unum,x,
     *      f,fhist,fctr,flag)
        end if

C       Unscale final x,f.
        call unscaleIF(n,x,u,l,x_orig)
        do 50 i=1,n
          x(i) = x_orig(i)
 50     continue

        f = f*fscale

C       Write out f history
        if (writ.ge.1) then
          write(mainout_unum,*) 'f history (unscaled)'
          if(writ.gt.10) then
            write(6,*) 'f history (unscaled)'
          end if
          do 60 i=1,fctr - 1
            write(mainout_unum,*) fhist(i)
            if(writ.gt.10) then
              write(6,*) fhist(i)
            end if
 60       continue
        end if

        if (writ.gt.10) then
          if(option(5).ne.1) then
            close(unit = mainout_unum)
          end if
        end if
        close(unit = ptsout_unum)



      end


C-----------------------------------------------------------------
C 1) inputsIF
C
C     This routine simply prints out what you've inputed to IFFCO.
C     Useful for a self-check.
C-----------------------------------------------------------------
      subroutine inputsIF(n,x,u,l,fscale,minh,maxh,maxit,
     *     restart,writ,termtol,f,ncuts,option,nprocs,
     *     mainout_unum,ptsout_unum)

        implicit none

C Arguments
        integer n, maxit(2), restart, writ, ncuts, nprocs, option(7)
        integer mainout_unum, ptsout_unum
        double precision x(n), u(n), l(n)
        double precision fscale, minh, maxh, termtol, f

C Local variables
        integer i, writbase

C       Calculate writbase
        if (writ.ge.10) then
          writbase = writ - 10
        else
          writbase = writ
        end if

C       OPEN OUTPUT FILES
        if((option(5).ne.1).and.(writbase.ne.0)) then
          open(unit = mainout_unum, 
     *      file = './iffco.out',
     *      status = 'unknown')
        else
          mainout_unum = option(6)
        end if

C       Open "points.out" file and create structure for info
        open(unit = ptsout_unum, 
     *    file = './points.out',
     *    status = 'unknown')
	
        if (writ.gt.0) then
          write(mainout_unum,*) 'fscale     = ',fscale
          write(mainout_unum,*) 'minh       = ',minh
          write(mainout_unum,*) 'maxh       = ',maxh
          write(mainout_unum,*) 'n          = ',n
          write(mainout_unum,*) 'maxit(1)   = ',maxit(1)
          write(mainout_unum,*) 'maxit(2)   = ',maxit(2)
          write(mainout_unum,*) 'restart    = ',restart
          write(mainout_unum,*) 'writ       = ',writ
          write(mainout_unum,*) 'termtol    = ',termtol
          write(mainout_unum,*) 'maxcuts    = ',ncuts

          write(mainout_unum,10) (i,x(i),i=1,n)
          write(mainout_unum,20) (i,u(i),i=1,n)
          write(mainout_unum,30) (i,l(i),i=1,n)
          write(mainout_unum,40) (i,option(i),i=1,7)
          write(mainout_unum,50) (nprocs)
        end if

        if (writ.gt.10) then
          write(6,*) 'fscale     = ',fscale
          write(6,*) 'minh       = ',minh
          write(6,*) 'maxh       = ',maxh
          write(6,*) 'n          = ',n
          write(6,*) 'maxit(1)   = ',maxit(1)
          write(6,*) 'maxit(2)   = ',maxit(2)
          write(6,*) 'restart    = ',restart
          write(6,*) 'writ       = ',writ
          write(6,*) 'termtol    = ',termtol
          write(6,*) 'ncuts      = ',ncuts

          write(6,10) (i,x(i),i=1,n)
          write(6,20) (i,u(i),i=1,n)
          write(6,30) (i,l(i),i=1,n)
          write(6,40) (i,option(i),i=1,7)
          write(6,50) (nprocs)
        end if

 10     format(' x(',i3,')     = ',e17.10)
 20     format(' u(',i3,')     = ',e17.10)
 30     format(' l(',i3,')     = ',e17.10)
 40     format(' option(',i3,')= ',i3)
 50     format(' running IFFCO on ',i3,' processors')

      end


C-----------------------------------------------------------------
C 2) initIF
C
C This routine initializes various variables and checks that
C the input variables have reasonable values.
C Output variables:
C       x_orig
C       h
C       quasi
C       flag: > 0 failure.
C
C-----------------------------------------------------------------
      subroutine initIF(n,n2,x,u,l,func,fscale,minh,maxh,maxit,restart,
     *   writ,termtol,ncuts,option,x_orig,h,quasi,flag,mainout_unum)

        implicit none

C Arguments
        integer n, n2, maxit(2), restart, writ, ncuts, quasi
        integer flag, option(7), mainout_unum
        double precision x(n), u(n), l(n)
        double precision fscale, minh, maxh, termtol
        double precision x_orig(n),h
        external func

C Local variables
        integer i, nerrs, mx, writbase

        parameter(mx = 24)
C Global variables
        integer nevalsIF
        double precision mepsIF,fminIF,xminIF(mx),fmaxIF
        common /globalIF/nevalsIF
        common /globalIF2/fminIF,xminIF,fmaxIF
        common /globalIF3/mepsIF


C       Calculate writbase
        if (writ.ge.10) then
          writbase = writ - 10
        else
          writbase = writ
        end if

C       Initialize variables.
        nerrs = 0
        nevalsIF = 0
        mepsIF = 1D-12
        quasi = option(1)

C       Make sure n2 >= n.
        if (n2.lt.n) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Error: Parameter mx < n.'
            write(mainout_unum,*) 'Make sure to change mx value in      
     *          all subroutines.'
          end if
          write(6,*) 'Error: Parameter mx < n.'
          write(6,*) 'Make sure to change mx value in all subroutines.'
          nerrs = nerrs + 1
        end if

C       Check consistency of upper bounds, lower bounds, and x.
        do 10 i=1,n
          if (u(i).le.l(i)) then
            if(writ.gt.0) then
              write(mainout_unum,20) i,i
            end if
            write(6,20) i,i
 20         format(' u(',i3,') is less than l(',i3,')')
            nerrs = nerrs + 1
          end if
          if (x(i).gt.u(i).or.x(i).lt.l(i)) then
            if(writ.gt.0) then
              write(mainout_unum,30) i
            end if
            write(6,30) i
 30         format(' x(',i3,') is out of bounds')
            nerrs = nerrs + 1
          end if
 10     continue


C       Check fscale non-negative
        if(fscale.lt.0.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'fscale negative'
          end if
          write(6,*) 'fscale negative'
          nerrs = nerrs + 1
        else if (fscale .eq. 0.0) then
          fscale=1.0
        end if


C       Check consistency of minh and maxh and that maxh <= .5.
        if (minh.gt.maxh) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'minh > maxh'
          end if
          write(6,*) 'minh > maxh'
          nerrs = nerrs + 1
        else if (maxh.gt.5D-1) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'maxh > .5'
          end if
          write(6,*) 'maxh > .5'
          nerrs = nerrs + 1
        else
          h = maxh
        end if


C       Check n > 0.
        if (n.le.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'n <= 0'
          end if
          write(6,*) 'n <= 0'
          nerrs = nerrs + 1
        end if


C       Check maxit(1),maxit(2) > 0.
        if (maxit(1).lt.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'maxit(1) < 0'
          end if
          write(6,*) 'maxit(1) < 0'
          nerrs = nerrs + 1
        else if (maxit(1).eq.0) then
          maxit(1) = 100
        end if
	
        if (maxit(2).lt.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'maxit(2) < 0'
          end if
          write(6,*) 'maxit(2) < 0'
          nerrs = nerrs + 1
	    else if (maxit(2).eq.0) then
	      maxit(2) = 100*n*n
	    end if

        if (restart.lt.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'restart < 0'
          end if
          write(6,*) 'restart < 0'
          nerrs = nerrs + 1
        end if

C       Check termtol > 0.
        if (termtol.lt.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'termtol < 0'
          end if
          write(6,*) 'termtol < 0'
          nerrs = nerrs + 1
        else if (termtol.eq.0d0) then
          termtol = 1.d0
        end if


C       Check ncuts >= 0.
        if (ncuts.lt.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'maxcuts < 0'
          end if
          write(6,*) 'maxcuts < 0'
          nerrs = nerrs + 1
        else if (ncuts.eq.0) then
          ncuts = 3
        end if

C       Check that option has legal values.
        if(option(1).lt.0 .or. option(1).gt.2) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Invalid value for option(1)'
          end if
          write(6,*) 'Invalid value for option(1)'
          nerrs = nerrs + 1
        end if
      
        if(option(2).lt.0 .or. option(2).gt.3) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Invalid value for option(2)'
          end if
          write(6,*) 'Invalid value for option(2)'
          nerrs = nerrs + 1
        end if

        if(option(3).lt.0 .or. option(3).gt.1) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Invalid value for option(3)'
          end if
          write(6,*) 'Invalid value for option(3)'
          nerrs = nerrs + 1
        end if

        if(option(4).lt.0 .or. option(4).gt.2) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Invalid value for option(4)'
          end if
          write(6,*) 'Invalid value for option(4)'
          nerrs = nerrs + 1
        end if

        if(option(5).lt.0 .or. option(5).gt.1) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Invalid value for option(5)'
          end if
          write(6,*) 'Invalid value for option(5)'
          nerrs = nerrs + 1
        end if

        if(option(6).lt.0) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Invalid value for option(6)'
          end if
          write(6,*) 'Invalid value for option(6)'
          nerrs = nerrs + 1
        end if

        if(option(7).lt.0 .or. option(7).gt.1) then
          if(writ.gt.0) then
            write(mainout_unum,*) 'Invalid value for option(7)'
          end if
          write(6,*) 'Invalid value for option(7)'
          nerrs = nerrs + 1
        end if

        if(option(3).eq.1) then
C       Set x_orig to center of box.
          do 40 i=1,n
            x_orig(i) = (u(i) + l(i)) / 2D0
 40       continue
        else
C         Copy x to x_orig
          do 50 i=1,n
            x_orig(i) = x(i)
 50       continue
        end if

        if (writ.ge.1) then
          write(mainout_unum,*) 
     *      '---------------------------------------------'
          write(mainout_unum,60) nerrs
          write(mainout_unum,*) 
     *      '---------------------------------------------'
        end if
        if(writ.ge.11) then
          write(6,*) '---------------------------------------------'
          write(6,60) nerrs
          write(6,*) '---------------------------------------------'
        end if

C       Since the "standard info" output lines don't have junk (num funcs,
C       current iterate) stuck in between them, the header is printed just
C       once, at the top.  This needs to be done after the number of errs line. 
        if(writbase.eq.1 .or. writbase.eq.3) then
          write(mainout_unum,*) ' m     ||x||        f          ||g||',
     *                 '         h        cuts'
          if(writ.gt.10) then
            write(6,*) ' m     ||x||        f          ||g||',
     *                 '         h        cuts'
          end if
        end if


 60       format(' There are a total of ',i3,' input errors')

        flag = nerrs

      end


C-----------------------------------------------------------------
C 3) statsIF
C
C This routine logs various statistics about the current
C iterate of IFFCO to iffco.out.
C
C writ = 1 - standard info
C writ = 2 - standard info + x
C writ = 3 - standard info, unscaled
C writ = 4 - standard info + x, unscaled
C Add 10 to print the info to stdout.
C The user has the option to print out scaled f,x or unscaled f,x.
C However \| x \| and \| GF \| will always refer to the scaled
C value.
C-----------------------------------------------------------------
      subroutine statsIF(n,x,u,l,fscale,h,f,writ,ncuts,g,its,unit)
        implicit none

C Arguments
        integer n, writ, ncuts, its, mx, unit
        double precision x(n), u(n), l(n), fscale
        double precision h, f, g(n)

        parameter(mx = 24)
C Local variables
        integer i, writbase
        double precision f_value, x_value(mx), nmx, nmg

C Global variables
        integer nevalsIF
        common /globalIF/ nevalsIF

        if (writ.ge.10) then
          writbase = writ - 10
        else
          writbase = writ
        end if

C       Copy f,x to f_value,x_value
        f_value = f
        do 10 i=1,n
          x_value(i) = x(i)
 10     continue

C       Compute norm(g) and norm(x)
        nmx = 0.d0
        nmg = 0.d0
        do 20 i=1,n
          nmx = nmx + x(i)*x(i)
          nmg = nmg + g(i)*g(i)
 20     continue
        nmx = dsqrt(nmx / dble(n))
        nmg = dsqrt(nmg / dble(n))

        if(writbase.eq.2.or.writbase.eq.4) then
          write(unit,*) ' m     ||x||        f          ||g||',
     *                 '         h        cuts'
          if (writ.gt.10) then
             write(6,*) ' m     ||x||        f          ||g||',
     *                 '         h        cuts'
          end if
        end if

C       Unscale the results.
        if(writbase.eq.3.or.writbase.eq.4) then
          f_value = f*fscale
          call unscaleIF(n,x,u,l,x_value)
        end if

C       Print out the statistics.
        if(writbase.ge.1) then
          if(ncuts.eq.-1) then
            write(unit,200) its,nmx,f_value, nmg, h
            if (writ.gt.10) then
              write(6,200) its,nmx,f_value, nmg, h
            end if
          else if(ncuts.eq.-2) then
            write(unit,210) its,nmx,f_value, nmg, h
            if (writ.gt.10) then
              write(6,210) its,nmx,f_value, nmg, h
            end if
          else if(ncuts.eq.-3) then
            write(unit,230) its,nmx,f_value, nmg, h
            if (writ.gt.10) then
              write(6,230) its,nmx,f_value, nmg, h
            end if
          else if(ncuts.eq.-4) then
            write(unit,240) its,nmx,f_value, nmg, h
            if (writ.gt.10) then
              write(6,240) its,nmx,f_value, nmg, h
            end if
	  else if(ncuts.eq.-5) then
            write(unit,250) its,nmx,f_value, nmg, h
            if (writ.gt.10) then
              write(6,250) its,nmx,f_value, nmg, h
            end if
          else if(ncuts.ge.0) then
            write(unit,220) its,nmx,f_value, nmg, h, ncuts
            if (writ.gt.10) then
              write(6,220) its,nmx,f_value, nmg, h, ncuts
            end if
          end if
        end if

C       Print out x components as well.
        if(writbase.eq.2.or.writbase.eq.4) then
          write(unit,30) (i,x_value(i),i=1,n)
          if (writ.gt.10) then
            write(6,30) (i,x_value(i),i=1,n)
          end if
 30       format(' x(',i3,')     = ',e17.10)

          write(unit,100) nevalsIF
          if(writ.gt.10) then
            write(6,100) nevalsIF
          end if

        end if


 100    format('Total function evaluations: ',i5)

 200    format(i3,' ', e10.4,' ',e10.4,' ',e10.4,' ',e10.4,
     *             '  Line search failure')
 210    format(i3,' ', e10.4,' ',e10.4,' ',e10.4,' ',e10.4,
     *             '  Convergence')
 220    format(i3,' ', e10.4,' ',e10.4,' ',e10.4,' ',e10.4,'   ',i3)

 230    format(i3,' ', e10.4,' ',e10.4,' ',e10.4,' ',e10.4,
     *             '  Maxit(1) exceeded')
 240    format(i3,' ', e10.4,' ',e10.4,' ',e10.4,' ',e10.4,
     *             '  Stencil Failure')
 250    format(i3,' ', e10.4,' ',e10.4,' ',e10.4,' ',e10.4,
     *             '  Maxit(2) exceeded')

      end


C-----------------------------------------------------------------
C 4) scaleIF
C
C This routine scales x_orig from the domain defined by l,u to the
C unit box.
C
C Output variables:
C       x_scal
C-----------------------------------------------------------------
      subroutine scaleIF(n,x_orig,u,l,x_scal)

        implicit none

C Arguments
        integer n
        double precision x_orig(n), u(n), l(n), x_scal(n)

C Local variables
        integer i

        do 10 i=1,n
          x_scal(i) = (x_orig(i) - l(i) ) / ( u(i) - l(i) )
 10     continue
      end


C-----------------------------------------------------------------
C 5) unscaleIF
C
C This routine unscales x_scal from the unit box to the domain defined
C by l,u.
C
C Output variables:
C       x_orig
C-----------------------------------------------------------------
      subroutine unscaleIF(n,x_scal,u,l,x_orig)

        implicit none

C Arguments
        integer n
        double precision x_orig(n), u(n), l(n), x_scal(n)

C Local variables
        integer i

        do 10 i=1,n
          x_orig(i) = ( u(i) - l(i) ) * x_scal(i) + l(i)
 10     continue
      end


C-----------------------------------------------------------------
C 6) funcIF
C
C This routine evaluates the objective function, 'func'.
C 'func' is expected to return flag = 0 for success.
C                                   > 0 for failure.
C
C Input variables:
C	x - scaled x value.
C Ouput variables:
C       f - scaled objective function value at x.
C-----------------------------------------------------------------
      subroutine funcIF(n,x,u,l,func,fscale,f,fflag, 
     *  writ, mainout_unum, ptsout_unum,
     *  idata, ilen, ddata, dlen, cdata, clen)

        implicit none

C Arguments
        integer n, writ, mainout_unum, ptsout_unum, fflag
        double precision x(n), u(n), l(n), fscale, f
        external func
        integer ilen, dlen, clen
        integer idata(ilen) 
        double precision ddata(dlen)
        character cdata(clen)

C Local variables
        integer mx, flag
        parameter(mx=24)
        double precision x_unscal(mx)

C Global variables
        integer nevalsIF

C Function Variables (global)
        common /globalIF/ nevalsIF
        
        call unscaleIF(n,x,u,l,x_unscal)

C       Initialize flag to success.
        flag = 0

C Call to outside function.  

        call func(n,x_unscal,f,flag, idata, ilen, ddata, dlen, 
     *    cdata, clen)

        fflag = flag
        if(flag.gt.0) then
          f = fscale
          if(writ.gt.0) then
            write(mainout_unum,*) 'function failed to evaluate!'
            write(mainout_unum,*) 'Setting f := fscale.'
          end if
          if(writ.gt.10) then
            write(6,*) 'function failed to evaluate!'
            write(6,*) 'Setting f := fscale.'
          end if
        end if

C       Check if current point is a minimum so far (unscaled).
C       fminIF unscaled because restarts change fscale.
        call minIF(n,x_unscal,f,flag)

C       Scale f value
        f = f / fscale

        nevalsIF = nevalsIF + 1

        call maxIF(f,flag)
        call pointsIF(n,x,f,u,l,fscale,ptsout_unum)
	
      end



C For the serial version, use this gradient routine.
C-----------------------------------------------------------------
C 7a) ser_gradIF
C
C This routine evaluates the gradient using a finite difference
C formula with a difference step of h.
C
C Ouput variables
C       g - gradient
C-----------------------------------------------------------------
      subroutine ser_gradIF(n,x,u,l,func,fscale,f,h,stencil,g,
     *  writ, mainout_unum, ptsout_unum,
     *  idata, ilen, ddata, dlen, cdata, clen)

        implicit none

C Arguments
        integer n, stencil, writ, mainout_unum, ptsout_unum
        double precision x(n), u(n), l(n), fscale, f, h, g(n)
        external func
        integer ilen, dlen, clen
        integer idata(ilen) 
        double precision ddata(dlen)
        character cdata(clen)

C Local variables
        integer mx
        parameter(mx = 24)
        integer i, fflagfor(mx), fflagback(mx), st(mx)
        double precision x2(mx), ffor(mx), fback(mx), minf, maxf

C       Initialize variables
        minf = f
        maxf = f
        stencil = 0
        do 10 i=1,n
          x2(i) = x(i)
 10     continue

C       See what type of difference gradient we can do for each direction.
        do 20 i=1,n
          st(i) = 0
          if (x(i)-h.ge.0D0)then
C           Able to do backward differences.
            st(i) = st(i) + 1
          end if
          if (x(i)+h.le.1D0)then
C           Able to do forward differences.
            st(i) = st(i) + 2
          end if
 20     continue

C       Get all function values needed to calculate the gradient.
        do 30 i=1,n
C         Backward differences            
          if (st(i).eq.1 .or. st(i).eq.3) then
            x2(i) = x2(i) - h
            call funcIF(n,x2,u,l,func,fscale,fback(i),
     *        fflagback(i),writ,mainout_unum,ptsout_unum,
     *        idata, ilen, ddata, dlen, cdata, clen)

C           Reset x2(i)
            x2(i) = x(i)

C           Check for extremal f values
            if (fback(i).lt.minf .and. fflagback(i).eq.0) then
              minf = fback(i)
            else if(fback(i).gt.maxf .and. fflagback(i).eq.0) then
              maxf = fback(i)
            end if 

	  end if

C         Forward differences
          if (st(i).eq.2 .or. st(i).eq.3) then
            x2(i) = x2(i) + h
            call funcIF(n,x2,u,l,func,fscale,ffor(i),
     *        fflagfor(i),writ,mainout_unum,ptsout_unum,
     *        idata, ilen, ddata, dlen, cdata, clen)

C           Reset x2(i)
            x2(i) = x(i)

C           Check for extremal f values
            if (ffor(i).lt.minf .and. fflagfor(i).eq.0) then
              minf = ffor(i)
            else if(ffor(i).gt.maxf .and. fflagfor(i).eq.0) then
              maxf = ffor(i)
            end if
            
	  end if
 30     continue

C       Set all infeasible function values to a value slightly
C       larger than the largest feasible value.
        maxf = maxf + 1.D-6 * abs(maxf)
        do 40 i = 1,n
          if(st(i).eq.1 .or. st(i).eq.3) then
            if(fflagback(i).gt.0) then
              fback(i) = maxf
            end if
          end if
          if(st(i).eq.2 .or. st(i).eq.3) then
            if(fflagfor(i).gt.0) then
              ffor(i) = maxf
            end if
          end if
 40     continue

C       Calculate the gradient
        do 50 i = 1,n
          if(st(i).eq.1) then
            g(i) = (f - fback(i)) / h
          else if(st(i).eq.2) then
            g(i) = (ffor(i) - f) / h
          else
            g(i) = (ffor(i) - fback(i)) / (2*h)
          end if
 50     continue

C       Check for stencil failure.
        if (f.le.minf) then
          stencil = 1
        end if

      end





C-----------------------------------------------------------------
C 8) minIF
C
C This routine checks to see if the point given has a lower function
C value than the current minimum function value.  If so it saves the
C point and value and returns flag = 1.  If not then it does
C nothing and returns flag = 0.
C Input variables:
C       n -- number of dimensions
C       x -- point, *unscaled*
C       f -- function value, *unscaled*
C Output variables:
C	flag
C-----------------------------------------------------------------
      subroutine minIF(n,x,f,flag)
        implicit none

C Arguments
        integer n, flag
        double precision x(n), f

C Local variables
        integer mx
        parameter(mx = 24)
        integer i

C Global variables
        double precision mepsIF, fminIF, xminIF(mx),fmaxIF
        common /globalIF2/fminIF,xminIF,fmaxIF
        common /globalIF3/mepsIF

        if (f.lt.fminIF) then
          fminIF = f
          do 10 i = 1,n
            xminIF(i) = x(i)
 10       continue
          flag = 1

        else
          flag = 0
        end if

      end


C-----------------------------------------------------------------
C 9) maxIF
C
C This routine checks to see if the point given has a greater function
C value then the current max function value.  If so it saves the
C value and returns flag = 1.  If not then it does nothing and
C returns flag = 0.
C Output variables:
C       flag
C-----------------------------------------------------------------
      subroutine maxIF(f,flag)

        implicit none

C Arguments
        integer flag
        double precision f

C Local variables
        integer mx
        parameter(mx = 24)

C Global variables
        double precision mepsIF, fminIF, xminIF(mx), fmaxIF
        common /globalIF2/fminIF,xminIF, fmaxIF
        common /globalIF3/mepsIF

        if (f.gt.fmaxIF) then
          fmaxIF = f
          flag = 1
        else
          flag = 0
        end if

      end


C-----------------------------------------------------------------
C 10) updateIF
C
C This routine simply copies xp,gp,fp to x,g,f.
C Output variables:
C	x,g,f
C-----------------------------------------------------------------
      subroutine updateIF(n,xp,gp,fp,x,g,f)

        implicit none

C Arguments
        integer n
        double precision xp(n), gp(n), fp, x(n), g(n), f

C Local variables
        integer i

        do 10 i=1,n
          x(i) = xp(i)
          g(i) = gp(i)
 10     continue

        f = fp

      end


C For serial versions, use the following linesearch

C-----------------------------------------------------------------
C 11a) ser_linesearchIF
C
C This routine implements a quadratic and cubic line search.
C p = direction of step: - H*g.
C Output variables:
C       ncuts: -1 Failure.
C       fp
C       xp
C-----------------------------------------------------------------
      subroutine ser_linesearchIF(n,x,u,l,func,fscale,g,p,f,minh,
     *     maxcuts,ncuts,fp,xp,writ, mainout_unum, ptsout_unum,
     *     idata, ilen, ddata, dlen, cdata, clen)

        implicit none

C Arguments
        integer n, maxcuts, ncuts, writ, mainout_unum, ptsout_unum
        double precision x(n), u(n), l(n), fscale, g(n) 
        double precision p(n), f, minh, fp, xp(n)
        external func
        integer ilen, dlen, clen
        integer idata(ilen) 
        double precision ddata(dlen)
        character cdata(clen)

C Local Variables
        integer mx
        parameter(mx = 24)
        integer i, active, cubic, flag, fflag
        double precision alpha, oldalpha, dfz, f2p, nrm
        double precision suffdec, lp, l2p, sgma
        double precision z(mx), a11, a21, a12, a22, v1, v2
        double precision a, b
        double precision lastx(mx)

C Global variables
        double precision mepsIF,fminIF,xminIF(mx),fmaxIF
        common /globalIF2/fminIF,xminIF,fmaxIF
        common /globalIF3/mepsIF

        ncuts = 0
        flag = 0
        cubic = 0
        sgma = 1D-4
        alpha = 1D0


C       Form new trial point: xp = P[x + alpha p]
        do 10 i=1,n
          z(i) = x(i) + alpha * p(i)
 10     continue


C       Form xp = P[z]. (Project z onto the unit box)
        do 20 i=1,n
          if (z(i).le.0D0) then
            xp(i) = 0D0
          else if (z(i).ge.1D0) then
            xp(i) = 1D0
          else
            xp(i) = z(i)
          end if
 20     continue

C       Evaluate fp = f(xp).
        call funcIF(n,xp,u,l,func,fscale,fp,fflag,
     *    writ, mainout_unum, ptsout_unum,
     *    idata, ilen, ddata, dlen, cdata, clen)

        do 25 i=1,n
          lastx(i) = xp(i)
 25     continue

C       Form suffdec = sgma * g' * (xp - x)  and  dfz = g'*p.
        suffdec = 0D0
        dfz = 0D0
        do 30 i=1,n
          suffdec = suffdec + g(i) * (xp(i) - x(i))
          dfz = dfz + g(i) * p(i)
 30     continue
        suffdec = suffdec * sgma


C       Iterate until 1) fp is sufficiently smaller than f
C                 or  2) ncuts exceeds maxcuts
C                 or  3) flag = 1
        do while (fp.gt.f + suffdec.and.ncuts.lt.maxcuts.
     *            and.flag.eq.0) 

C         Find any new active constraints of xp.  
C                    active = 1         new active constraints
C                    active = 0         no new active constraints
          active = 0
          do 40 i=1,n
            z(i) = x(i) + alpha * p(i)
            if( (z(i).lt.0D0.and.x(i).gt.mepsIF).or.
     *          (z(i).gt.1D0.and.x(i).lt.1D0-mepsIF) ) then
              active = 1
            end if
 40       continue          

C         If there are new active constraints do not use quadratic
C         model.
          if(active.eq.1) then
            oldalpha = alpha
            alpha = 5D-1 * alpha
          else
C           Use quadratic model.
            if(cubic.eq.0) then
              oldalpha = alpha
              alpha = - alpha * dfz / (2*(fp - f - alpha*dfz))
              alpha = oldalpha * alpha
              cubic = 1
C           Use cubic model
            else if (cubic.eq.1) then
              lp = alpha
              l2p = oldalpha
              a11 = 1 / (lp*lp)
              a21 = - l2p / (lp*lp)
              a12 = - 1 / (l2p*l2p)
              a22 = lp / (l2p*l2p)

              v1 = fp - f - dfz*lp
              v2 = f2p - f - dfz*l2p
 
              a = 1 / (lp - l2p) * (a11 * v1 + a12 * v2)
              b = 1 / (lp - l2p) * (a21 * v1 + a22 * v2)

              oldalpha = alpha

              if(dabs(a).gt.mepsIF) then
                alpha = (-b + dsqrt(b*b - 3*a*dfz)) / (3*a)
              else
C             Cubic model failed.  Use quadratic model
                alpha = - alpha * dfz / (2*(fp - f - alpha*dfz))
                alpha = oldalpha * alpha
              end if
            end if
            f2p = fp
          end if


C         Bound alpha between [.1*oldalpha, .5*oldalpha].
          if (alpha.gt.(5D-1)*oldalpha) then
            alpha = (5D-1)*oldalpha
          else if (alpha.lt.(1D-1)*oldalpha) then
            alpha = (1D-1)*oldalpha 
          end if

C         Form new trial point: xp = P[x + alpha p]
          do 50 i=1,n
            z(i) = x(i) + alpha*p(i)
 50       continue

C         Form xp = P[z]. (Project z onto the unit box)
          do 60 i=1,n
            if (z(i).le.0D0) then
              xp(i) = 0D0
            else if (z(i).ge.1D0) then
              xp(i) = 1D0
            else
              xp(i) = z(i)
            end if
 60       continue

C         Check that the step length (alpha * p) has not
C         been reduced below .01*h.
C         Form nrm
          nrm = 0D0
          do 70 i=1,n
            a = alpha*p(i)
            nrm = nrm + a*a
 70       continue
          nrm = dsqrt(nrm / dble(n))

          if(nrm.lt.(1D-2)*minh) then
            flag = 1
          else
C           Check to see if line search came up w/ same point as last time
            i=1
            do while (dabs(xp(i)-lastx(i)).le.mepsIF.and.i.lt.n)
              i=i+1
            end do

            if(i.lt.n.or.dabs(xp(n)-lastx(n)).gt.mepsIF ) then
C             Evaluate fp = f(xp).
              call funcIF(n,xp,u,l,func,fscale,fp,fflag,
     *          writ, mainout_unum, ptsout_unum,
     *          idata, ilen, ddata, dlen, cdata, clen)

C             Save point.
              do 65 i=1,n
                lastx(i) = xp(i)
 65           continue
            end if
          end if


C         Form suffdec = sgma * g' * (xp - x) 
          suffdec = 0D0
          do 80 i=1,n
            suffdec = suffdec + g(i) * (xp(i) - x(i))
 80       continue
          suffdec = suffdec * sgma


          ncuts = ncuts + 1
          
        end do

        
C       Linesearch failure.  Return ncuts = -1
        if (fp.gt.f + suffdec) then
          ncuts = -1
C         Set fp = f (used for taking minimums)
          fp = f
        end if

      end



C For parallel versions, use this linesearch. 

C End #ifdef USE_SERIAL


C-----------------------------------------------------------------
C 12) eyeIF
C
C This routine sets the matrix B = I.
C Output variables:
C       B
C-----------------------------------------------------------------
      subroutine eyeIF(n,B)

        implicit none

C Arguments
        integer n
        double precision B(n,n)

C Local variables
        integer i,j

        do 10 i=1,n
          do 20 j=1,n
            if(i.eq.j) then
              B(i,j) = 1D0
            else
              B(i,j) = 0D0
            end if
 20       continue
 10     continue


      end


C-----------------------------------------------------------------
C 13) evaltolIF
C
C This routine evaluates the tolerance at the current point x.
C We define the tolerance by \| x - P[x - \grad_h f(x)] \|
C Output variables:
C       tol
C-----------------------------------------------------------------
      subroutine evaltolIF(n,x,u,l,fscale,g,tol)
        implicit none

C Arguments
        integer n
        double precision x(n), u(n), l(n), fscale
        double precision g(n), tol

C Local variables
        integer mx
        parameter(mx = 24)
        integer i
        double precision diff(mx), proj(mx)


C       Form P[x - \grad_h f(x)].
        do 10 i=1,n
          diff(i) = x(i) - g(i)

          if (diff(i).le.0D0) then
            proj(i) = 0.d0
          else if (diff(i).ge.1) then
            proj(i) = 1.d0
          else
            proj(i) = diff(i)
          end if
 10     continue

C       Compute \| x - P[x - \grad_h f(x)] \|
        tol = 0.d0
        do 20 i=1,n
          tol = tol + (x(i) - proj(i))*(x(i) - proj(i))
 20     continue

        tol = dsqrt(tol / dble(n))

      end


C-----------------------------------------------------------------
C 14) quasiIF
C
C This routine initializes and updates the quasi-Newton matrix.
C It handles either the SR1 or the BFGS updates.
C quasi = option(1)
C quasi = 0:  projected gradient alg.
C quasi = 1:  SR1 update
C quasi = 2:  BFGS update
C Output variables:
C       B
C-----------------------------------------------------------------
      subroutine quasiIF(n,x,xp,g,gp,quasi,option,B,
     *  writ,mainout_unum)
        implicit none

C Arguments
        integer n, quasi, option, writ, mainout_unum
        double precision x(n),xp(n),g(n),gp(n),B(n,n)

C Local variables
        integer mx
        parameter(mx = 24)
        integer i,j, xp_act(mx), x_act(mx), flag
        double precision y(mx), s(mx), C(mx,mx), D(mx,mx)
        double precision t(mx), r, r2

C Global variables
        double precision mepsIF,fminIF,xminIF(mx),fmaxIF
        common /globalIF2/fminIF,xminIF,fmaxIF
        common /globalIF3/mepsIF


C       Find active sets for x_+ and x.
        do 10 i=1,n
          if ( (xp(i).lt.mepsIF.and.gp(i).gt.0D0).or.
     *           (xp(i).gt.(1D0 - mepsIF).and.gp(i).lt.0D0) ) then
            xp_act(i) = 1
          else
            xp_act(i) = 0
          end if

          if ( (x(i).lt.mepsIF.and.g(i).gt.0D0).or.
     *           (x(i).gt.(1D0 - mepsIF).and.g(i).lt.0D0) ) then
            x_act(i) = 1
          else
            x_act(i) = 0
          end if

 10      continue

C       Form y = gf(x_+) - gf(x)  and  s = x_+ - x
        do 20 i=1,n
          y(i) = gp(i) - g(i)
          s(i) = xp(i) - x(i)
 20     continue


C       Check to see if xp_act = x_act.
        flag = 0
        do 30 i=1,n
          if (xp_act(i).ne.x_act(i)) then
            flag = 1
          end if
 30     continue

C       Form P_{I_c} B_c P_{I_c}
C       Zero out the rows and columns of B wrt x_act.
        do 40 i=1,n
          if ( x_act(i).eq.1) then
            do 50 j=1,n
              B(i,j) = 0D0
              B(j,i) = 0D0
 50         continue
          end if
 40     continue

C       Set B = I
        if(quasi.eq.0.or.(flag.eq.1.and.option.eq.1) ) then
          call eyeIF(n,B)

C SR1 update (B = approximate Hessian)
C   B = B + [(y - B*s) (y-B*s)^T] / (y - B*s)'*s
        else if(quasi.eq.1) then

C         Form (y - B*s).
          do 60 i=1,n
            t(i) = 0D0
            do 70 j=1,n
              t(i) = t(i) + B(i,j)*s(j)
 70         continue
            t(i) = y(i) - t(i)
 60       continue

C         Form (y - B*s)'*s.
          r = 0D0
          do 80 i=1,n
            r = r + t(i)*s(i)
 80       continue

C         if |(y - B*s)'*s| is too small then skip SR1 update.
          if(dabs(r).gt.mepsIF) then

C           Form [(y - B*s) (y-B*s)^T]
            do 90 i=1,n
              do 100 j=1,n
                C(i,j) = t(i)*t(j)
 100          continue
 90         continue

C           Update B
            do 110 i=1,n
              do 120 j=1,n
                B(i,j) = B(i,j) + C(i,j) / r
 120          continue
 110        continue
          end if


C BFGS update (B = approximate inverse Hessian)
C   B = (I-s*y'/y's)*B*(I-s*y'/y's) + s*s'/y'*s
        else if(quasi.eq.2) then
C         Form r = (s'*y)
          r = 0D0
          do 130 i=1,n
            r = r + s(i)*y(i)
 130      continue


C         Check that r is not <= 0.
          if(r.gt.0) then
C           Form t = B*y
            do 140 i=1,n
              t(i) = 0D0
              do 150 j=1,n
                t(i) = t(i) + B(i,j)*y(j)
 150          continue
 140        continue

C           Form r2 = (1 + y'*B*y / r )
            r2 = 0D0
            do 160 i=1,n
              r2 = r2 + t(i)*y(i)
 160        continue
            r2 = 1D0 + r2 / r


C           Form C = r2*(s*s')  and  D = B*y*s'
            do 170 i=1,n
              do 180 j=1,n
                C(i,j) = r2 * s(i) * s(j)
                D(i,j) = t(i) * s(j)
 180          continue
 170        continue

C           Update B
            do 190 i=1,n
              do 200 j=1,n
                B(i,j) = B(i,j) + ( C(i,j) - D(i,j) - D(j,i) ) / r
 200          continue
 190        continue
          else
            if(writ.gt.0) then
              write(mainout_unum,*) 'B not positive definite'
            end if
            if(writ.gt.10) then
              write(6,*) 'B not positive definite'
            end if 
            call eyeIF(n,B)
          end if


        end if


C       Form P_{I_+} B P_{I_+} + P_{A_+}.
        do 210 i=1,n
          if ( xp_act(i).eq.1) then
            do 220 j=1,n
              if(j.ne.i) then
                B(i,j) = 0D0
                B(j,i) = 0D0
              else
                B(i,i) = 1D0
              end if
 220        continue
          end if
 210    continue

      end


C-----------------------------------------------------------------
C 15) stepIF
C
C This routine forms the step: p = - H^{-1} g
C where g is the gradient at x and H is an approximation
C to the Hessian, \grad^2 f(x).
C
C For BFGS, B = H^{-1}
C For SR1, B = H
C We also set the condition that \| p / n \| >= 2*h.
C Output variables:
C	p
C
C-----------------------------------------------------------------
      subroutine stepIF(n,g,h,B,quasi,writ,mainout_unum,p)
        implicit none

C Arguments
        integer n, quasi, writ, mainout_unum
        double precision g(n), h, B(n,n), p(n)

C Local Variables
        integer mx
        parameter(mx = 24)
        integer i, j, jpvt(mx), info, writbase
        double precision C(mx,mx), work(mx), nrm

        if(writ.ge.10) then
          writbase = writ - 10
        else
          writbase = writ
        end if


        if(quasi.eq.1) then

C SR1 Case:
C         C := B
          do 10 i=1,n
            do 20 j=1,n
              C(i,j) = B(i,j)
 20         continue
 10       continue

C         Factor B (Note dchdc must use B because C has wrong dimensions)
          call dchdc(B,n,n,work,jpvt,0,info)

C         Set p = -g
          do 30 i=1,n
            p(i) = - g(i)
 30       continue

          if(info.lt.n) then
C           B isn't positive definite, set B = I  &  p = -g.
            if(writbase.gt.0) then
              write(mainout_unum,*) 'B not positive definite'
              if(writ.gt.10) then
                write(6,*) 'B not positive definite'
              end if
            end if
            call eyeIF(n,B)
          else
C           Set p = -g and call dposl (which will compute p = B^{-1} (-g) )
            call dposl(B,n,n,p)

C           B := C
            do 35 i=1,n
              do 40 j=1,n
                B(i,j) = C(i,j)
 40           continue
 35         continue

          end if

        else if(quasi.eq.2.or.quasi.eq.0) then

C BFGS Case:
C         Compute: p = - B * g
          do 50 i = 1,n
            p(i) = 0.D0
            do 60 j = 1,n
              p(i) = p(i) - B(i,j) * g(j)
 60         continue
 50       continue

        end if


C       Modify p so that \|p\| / n^1/2 >= 1*h.
C       Compute nrm = \|p\| / n^1/2
        nrm = 0.d0
        do 70 i=1,n
          nrm = nrm + p(i)*p(i)
 70     continue
        nrm = dsqrt(nrm / dble(n))

C       Allows at least one point to check in linesearch
        if (nrm.lt.1D0*h) then
          do 80 i=1,n
            p(i) = 1D0*h*p(i) / nrm
 80       continue
        end if



      end



C-----------------------------------------------------------------
C 16) restartIF
C
C This routine is called when a restart is to be done.  It
C reinitializes h, fscale, and fmaxIF.
C Input variables:
C       f
C       maxh
C       writ
C Output variables (excluding globals):
C       fscale
C       h
C
C-----------------------------------------------------------------
      subroutine restartIF(f,maxh,writ,mainout_unum,fscale,h)
        implicit none

C Arguments
        integer writ, mainout_unum
        double precision maxh, f, fscale, h

C Local variables
        integer mx, writbase
        parameter(mx = 24)
        double precision f_unscal

C Global variables
        double precision fminIF, xminIF(mx), mepsIF, fmaxIF
        common /globalIF2/fminIF,xminIF,fmaxIF
        common /globalIF3/mepsIF

        if (writ.gt.10) then
          writbase = writ - 10
        else
          writbase = writ
        end if

C       Printout fminIF
        if(writbase.eq.1.or.writbase.eq.2) then
          write(mainout_unum,*) 'fminIF = ',fminIF / fscale
          if(writ.gt.10) then
            write(6,*) 'fminIF = ',fminIF / fscale
          end if
        else if(writbase.eq.3.or.writbase.eq.4) then
          write(mainout_unum,*) 'fminIF = ',fminIF
          if(writ.gt.10) then
            write(6,*) 'fminIF = ',fminIF
          end if
        end if

C       Printout fmaxIF
        if(writbase.eq.1.or.writbase.eq.2) then
          write(mainout_unum,*) 'fmaxIF = ',fmaxIF
          if(writ.gt.10) then
            write(6,*) 'fmaxIF = ',fmaxIF
          end if
        else if(writbase.eq.3.or.writbase.eq.4) then
          write(mainout_unum,*) 'fmaxIF = ',fmaxIF * fscale
          if(writ.gt.10) then
            write(6,*) 'fmaxIF = ',fmaxIF * fscale
          end if
        end if

C       Reinitialize h
        h = maxh

C       Unscale f
        f_unscal = f * fscale

C       Set new fscale
        if(fmaxIF.gt.1D1.or.fmaxIF.lt.1D-1) then
          fscale = fmaxIF*fscale
          write(6,*) 'New fscale = ',fscale
        end if

C       Rescale f
        f = f_unscal / fscale

C       Reinitialize fmaxIF
        fmaxIF = f

      end


C-----------------------------------------------------------------
C 17) takeminIF
C
C This routine replaces the current point with the point stored
C in xminIF(n) if fminIF < f_unscaled(x).
C If so it records the minimim value in fhist.
C-----------------------------------------------------------------
      subroutine takeminIF(n,u,l,mx2,fscale,writ,unit,x,f,fhist,fctr,
     *                     flag)
        implicit none

C Arguments
        integer n, mx2, writ, fctr, unit, flag
        double precision u(n), l(n), fscale, x(n), f, fhist(mx2)

C Local variables
        integer mx
        parameter(mx = 24)
        integer i, writbase
        double precision f_value, x_value(mx)

C Global variables
        double precision fminIF, xminIF(mx), mepsIF, fmaxIF
        common /globalIF2/fminIF,xminIF,fmaxIF
        common /globalIF3/mepsIF

        flag = 0

C       Check if min f is less than unscaled f.
        if (fminIF.lt.f*fscale) then

          flag = 1
C         Update f,x w/ scaled values of fminIF and xminIF
          f = fminIF / fscale
          call scaleIF(n,xminIF,u,l,x)

          if (writ.gt.10) then
            writbase = writ - 10
          else
            writbase = writ
          end if


C         Copy f,x to f_value,x_value
          f_value = f
          do 10 i=1,n
            x_value(i) = x(i)
 10       continue

C         Unscale the results.
          if(writbase.eq.3.or.writbase.eq.4) then
            f_value = f*fscale
            call unscaleIF(n,x,u,l,x_value)
          end if


          if(writbase.ge.1) then
            write(unit,*) 'Minimum Taken: f = ',f_value
            if(writ.gt.10) then
              write(6,*) 'Minimum Taken: f  = ',f_value
            end if
          end if

          if(writbase.eq.2.or.writbase.eq.4) then
            write(unit,*)'unscaled x'
            write(unit,20) (i,x_value(i),i=1,n)
            if(writ.gt.10) then
              write(6,*)'unscaled x'
              write(6,20) (i,x_value(i),i=1,n)
            end if
          end if


C         Save f history
          fhist(fctr) = fminIF
          fctr = fctr + 1
        end if

 20     format(' x(',i3,')     = ',e17.10)


      end



	
C-----------------------------------------------------------------
C 18) pointsIF
C
C This routine will write out to the file "points.out". The 
C first column will be the function evaluation count, the
C second will be the the function value, and the next n
C columns will be the components of x (unscaled).
C-----------------------------------------------------------------
      subroutine pointsIF(n,x_orig,f,u,l,fscale,ptsout_unum)
        implicit none

C Arguments
        integer n, ptsout_unum
        double precision x_orig(n), f, u(n), l(n), fscale
	
C Local variables
        integer mx, i
        parameter(mx=24)
        double precision x(mx)

C Global variables
        integer nevalsIF
        common /globalIF/ nevalsIF
	
        call unscaleIF(n, x_orig, u, l, x)
        write(ptsout_unum,*) nevalsIF,f*fscale,(x(i),i=1,n)
	
      end



