C----------------------------------------------------------------- 
C Driver program for IFFCO test problems.
C This program runs IFFCO on a collection of 15 test problems and compares
C the results with expected results to determine if IFFCO has been
C installed and compiled properly.
C
C For each test problem, this program reads a configuration file to get
C the proper parameters to pass to IFFCO.
C 
C Users may want to use this file as the basis for their own IFFCO driver
C program.  There is also a simpler example of a driver program in 
C sample_main.f.
C----------------------------------------------------------------- 

      program main

      implicit none

C Declare local variables
      integer i, prob, n, actnevals, nfails
      character*2 fname
      character*10 probname
      double precision actmin, mintol, nevalstol
      double precision minerr, nevalserr

C Declare IFFCO global variables.  These contain information about
C the result of the run.
      integer nevalsIF
      double precision fminIF, xminIF(24), fmaxIF
      common /globalIF/nevalsIF
      common /globalIF2/fminIF,xminIF,fmaxIF


C Declare IFFCO parameters.
C     n is the largest dimension that will be used.
      parameter(n = 10)

      external myfunc
      double precision x(n), u(n), l(n), fscale, minh, maxh
      double precision termTol, f
      integer dim, maxit(2), restart, writ, maxcuts, option(7)

      integer ilen, dlen, clen
C     Define the size of any extra data to be passed here
      parameter(ilen = 2, dlen = 1, clen = 1)
      integer idata(ilen)
      double precision ddata(dlen)
      character cdata(clen)



C     Counter for the number of test problem failures.
      nfails = 0

C     These tolerances specify how close the test results should be to
C     the control results.
      mintol = 0.01
      nevalstol = 0.01

C     Open file for test output 

      open(unit = 34, file = 'logfile')


C     Loop through all the problems in the test suite.
      do 10 prob = 0,11 
        write(unit = fname, fmt='(I2)') prob + 10
        open(unit = 33, file = 'testdata/'//fname)

C       Initialize IFFCO parameters by reading from the ini file
C       Dimension of problem
        read(33, *) dim

C       Starting point
        do 20 i = 1, dim
           read(33, *) x(i)
 20     continue

C       Upper bounds
        do 30 i = 1, dim
          read(33, *) u(i)
 30     continue

C       Lower bounds
        do 40 i = 1, dim
          read(33, *) l(i)
 40     continue

C       Scaling
        read(33, *) fscale

C       Smallest and largest scales
        read(33, *) minh
        read(33, *) maxh

C       Max iterations on a given scale and for entire algorithm
        read(33, *) maxit(1)
        read(33, *) maxit(2)

C       Number of restarts to do
        read(33, *) restart

C       Level of output
        read(33, *) writ

C       Termination tolerance
        read(33, *) termTol

C       Maximum cuts to take in a linesearch
        read(33, *) maxcuts

C       Options
        read(33, *) option(1)
        read(33, *) option(2)
        read(33, *) option(3)
        read(33, *) option(4)
        read(33, *) option(5)
        option(6) = 0
        option(7) = 0

C       Set additional, problem-specific data, if necessary.
        if ((prob .ge. 4) .and. (prob .le. 6)) then
           idata(1) = 5
           if (prob .eq. 5) then
              idata(1) = 7
           else if (prob .eq. 6) then
              idata(1) = 10
           end if
        else if ((prob .ge. 7) .and. (prob .le. 8)) then
C          dim = 3
           idata(1) = 4
        end if


C       Set the number of the test problem to run
        idata(ilen) = prob
C       Start IFFCO
        call iffco(myfunc,x,u,l,fscale,minh,maxh,dim,maxit,
     *       restart,writ,termTol,f,maxcuts,option, 
     *       idata, ilen, ddata, dlen, cdata, clen)

C       Get the problem's name from ini file
        read(33,*) probname

C       Get control results from ini file
        read(33,*) actmin
        read(33,*) actnevals


C       Close the problem's ini file.
        close(unit = 33)

C       Print comparison of the result with the control version 

        write(34,*) 'Problem ', prob, ': ', probname
        write(34,*)  '        Current   Control   ' 
        write(34,*)  '----------------------------'
        write(34,60) 'Min:    ', f, actmin
        write(34,70) 'Fevals: ', nevalsIF, actnevals
        write(34,*) ' '
        write(34,*) ' '

 60     format(a, f10.6, ' ', f10.6)
 70     format(a, i10, ' ', i10) 

C       Figure out whether this problem 'passed' or not.
        nevalserr = abs(nevalsIF - actnevals)
        nevalserr = nevalserr/abs(actnevals)

        minerr = abs(f - actmin)
        if(actmin .ne. 0) then
          minerr = minerr/abs(actmin)
        end if

        if((nevalserr.le.nevalstol).and.
     *     (minerr.le.mintol)) then
          write(*,*) 'Problem ', prob, ': ', probname, ' worked.' 
        else
          write(*,*) 'Problem ', prob, ': ', probname, ' failed.'
          nfails = nfails + 1
        end if

 10   continue

C     Close the log file and print a summary for all problems

      close(unit = 34)
      write(*,*) ' '
      write(*,*) ' '
      if(nfails.eq.0) then
        write(*,*) 'All test problems worked. '
      else
        write(*,*) nfails, ' test problems failed.'
      end if

      end
